/************************************************************************************
* This program is licensed under the BSD two-clause license, a copy of which
* is included below.
*
* Copyright (c) 2012, Claudio D. Lapilli
* All rights reserved.
*
* Redistribution and use in source and binary forms, with or without modification,
* are permitted provided that the following conditions are met:
*
* x-Redistributions of source code must retain the above copyright notice, this list
*   of conditions and the following disclaimer.
* x-Redistributions in binary form must reproduce the above copyright notice, this
*   list of conditions and the following disclaimer in the documentation and/or other
*   materials provided with the distribution.
*
* THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY
* EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
* OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT
* SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
* INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
* TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
* BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
* CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
* ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
* DAMAGE.
**************************************************************************************/








#include <stdlib.h>

#include "cleanfs.h"

#define FSDEBUG

#include "implement.h"

// FORMAT THE GIVEN DISK, CREATING A NEW FILESYSTEM USING AT MOST NumSectors
// USING THE SPECIFIED ClusterSize

int FSFormatVolume(struct FS_DISK *Disk,clfs_64bit NumSectors,clfs_32bit ClusterSize,clfs_char *VolLabel)
{
struct FS_VOLUME fs;
clfs_32bit tmp;
clfs_byte *buffer;
clfs_32bit f;
clfs_64bit cluster;
clfs_64bit dirsize;

// FORCE CLUSTERSIZE TO BE A POWER OF 2
tmp=1;
while(tmp<=ClusterSize) tmp<<=1;

ClusterSize=tmp>>1;

// FILL VOLUME STRUCTURE

FSMemcpy(&(fs.Disk),Disk,sizeof(struct FS_DISK));

fs.ClusterSize=ClusterSize;
fs.ClusterBits=1;
while( (1<<fs.ClusterBits)<fs.ClusterSize) ++fs.ClusterBits;

fs.SectorsPerCluster=(ClusterSize+Disk->SectorSize-1)/Disk->SectorSize;
fs.NumClusters=(NumSectors)/fs.SectorsPerCluster;
fs.Flags=0; // TO BE DEFINED
fs.VolName=VolLabel;

buffer=(clfs_byte *)FSMalloc(ClusterSize);
if(!buffer) return FS_NOMEM;

struct FS_BOOTRECORD *boot=(struct FS_BOOTRECORD *)buffer;

// WRITE BOOT SECTOR INFORMATION


boot->EntryType=DE_BOOTRECORD;
boot->Magic=FSBOOTRECORD_MAGIC;     // "CLFS" AS STORED ON-DISK
boot->BytesPerCluster=ClusterSize;
boot->TotalClusters=fs.NumClusters;
boot->RootDirCluster=0; // ROOT DIR HAS BOOT SECTOR EMBEDDED

clfs_char *src,*dest;

dest=boot->VolName;
src=VolLabel;
for(f=0;f<36;++f)
{
    *dest=*src;
    if(*src) ++src;
    ++dest;
}

dirsize=DIRENTRY_SIZE;

struct FS_NAMERECORD *namer;
namer=(struct FS_NAMERECORD *)boot+1;
while(*src) {
namer->EntryType=DE_NAME;
dest=namer->Name;
for(f=0;f<NAMERECORD_LENGTH;++f) {
    *dest=*src;
    if(*src) ++src;
    ++dest;
}
namer++;
dirsize+=DIRENTRY_SIZE;
if( ((clfs_byte *)namer)>=buffer+Disk->SectorSize) {
    // BOOT RECORD DOESN'T FIT IN A SINGLE SECTOR!
    // TRUNCATE VOLUME NAME RIGHT HERE
    break;
}
}

cluster=0;

// DONE WITH BOOT RECORD, NOW CONTINUE CREATING EMPTY ROOT DIRECTORY IMMEDIATELY AFTER
struct FS_FILERECORD *frec=(struct FS_FILERECORD *)namer;

if( ((clfs_byte *)frec)>=buffer+ClusterSize) {
    // ROOT DIRECTORY DOESN'T FIT W/BOOT RECORD, MOVE TO NEXT CLUSTER
    boot->RootDirCluster=1;
    cluster=1;
    dirsize=0;
    frec=(struct FS_FILERECORD *)buffer;
}

// WRITE FIRST CLUSTER NOW WITH BOOT RECORD
FSWriteCluster(0,buffer,1,&fs);

frec->StartCluster=cluster;
frec->EntryType=DE_FILE;
frec->Attributes=FSATTR_DIR;
frec->FileSize=dirsize+2*DIRENTRY_SIZE;         // WHATEVER WAS TAKEN BY THE BOOT RECORD, PLUS THE '.' ENTRY AND THE FREE SPACE BITMAP ENTRY.
frec->CreateDate=frec->ModifiedDate=FSTimeDateNow();
frec->NumClusters=1;
frec->ExtendedDates=0;
FSMemset(frec->Name,0,DIRENTRY_NAME_LENGTH);
frec->Name[0]='.';
frec->FoldedCRC=FSCalcFoldedNameCRC(frec->Name);
frec->NameCRC=FSCalcNameCRC(frec->Name);


++frec;


if( ((clfs_byte *)frec)>=buffer+ClusterSize) {
    // INCREASE THE NUMBER OF CLUSTERS TO TWO
    (--frec)->NumClusters++;
    FSWriteCluster(cluster,buffer,1,&fs);
    ++cluster;
    frec=(struct FS_FILERECORD *)buffer;
}

// NOW CREATE THE FREE CLUSTERS BITMAP

// TODO: MOVE FRAGMENTS OF THE BITMAP THROUGHOUT THE VOLUME TO IMPROVE LOCALITY

frec->EntryType=DE_FILE;
frec->Attributes=FSATTR_HIDDEN | FSATTR_IMMUTABLE;
frec->CreateDate=frec->ModifiedDate=FSTimeDateNow();
frec->FileSize= (fs.NumClusters+7)>>3;
frec->FileSize=(frec->FileSize+(fs.ClusterSize-1))&(~(fs.ClusterSize-1));
clfs_char *frname=FSSTRING_FREECLUSTERS;
for(f=0;f<DIRENTRY_NAME_LENGTH;++f)
{
    frec->Name[f]=*frname;
    if(*frname) ++frname;
}

frec->StartCluster=cluster+1;   // START RIGHT AFTER THE ROOT DIR
tmp=frec->NumClusters=frec->FileSize>>fs.ClusterBits;   // SAVE THE NUMBER OF CLUSTERS IN tmp TO FILL LATER

++frec;

// CLEAR THE REST OF THE CLUSTER
FSMemset(frec,0,ClusterSize-((clfs_byte *)frec-buffer));

// WRITE THE CLUSTER
FSWriteCluster(cluster,buffer,1,&fs);

// CLEAR THE ENTIRE BUFFER TO WRITE EMPTY SPACE BITMAP
FSMemset(buffer,0,fs.ClusterSize);


for(f=tmp-1;f>0;--f)    // LOOP THROUGH ALL CLUSTERS, FROM LAST TO FIRST
{
    FSWriteCluster(cluster+1+f,buffer,1,&fs);
}

buffer[0]=1;    // SET DIRTY BIT, FREE SPACE WILL BE SCANNED ON FIRST MOUNT

FSWriteCluster(cluster+1,buffer,1,&fs);


FSFree(buffer);
return FS_OK;
}

// DO A DISKCHECK ON A SINGLE ENTRY
int FSCheckFileEntry(struct FS_FILE *entry)
{
    // CHECK FILE ATTRIBUTES
    // 1-UNSUPPORTED ATTRIBUTES
    if(entry->F.Attributes&(~(FSATTR_DIR|FSATTR_CASEINSENS|FSATTR_DIRCASEINSENS|FSATTR_IMMUTABLE|FSATTR_HIDDEN))) {
        // FILE USES UNSUPPORTED ATTRIBUTES
    }

    // CHECK DATES

    // 0-CREATION AND MODIFY DATES MUST BE >0
    // 1-CREATION DATE MUST BE < CURRENT TIME
    // 2-MODIFY DATE MUST BE >= CREATION DATE
    // 3-MODIFY DATE MUST BE < CURRENT TIME


    // CHECK FILE SIZE MUST MATCH THE STORAGE ALLOCATED

    // CHECK NAME CRC AND FOLDED NAME CRC ARE CORRECT

    // CHECK ALLOCATED CLUSTERS

    // 1-FRAGMENTS DON'T OVERLAP
    // 2-ALL FRAGMENTS ARE MARKED AS USED IN THE FREE SPACE BITMAP
    // 3-FRAGMENTS ARE NOT OCCUPIED BY OTHER FILES (CROSS LINKED)

    // CHECK FREE SPACE BITMAP CONSISTENCY
    // 1-FIND CROSS-LINKED FILES
    // 2-FIND LOST CLUSTERS

}
